import AlteryxPythonSDK as Sdk
import connectors_core.alteryx as alteryx
import connectors_core.alteryx.helpers as alteryx_helpers

from connectors_core.bigquery import BigQuerySchemaReader
from .common import (BIG_QUERY_SCOPE,
                     OAUTH_AUD,
                     CannotInferRepeatedRecordLimit,
                     DEFAULT_MAX_FIELD_LENGTH,
                     BIG_QUERY_API_BASE)

from google.oauth2 import service_account, credentials
from google.cloud import bigquery
from google.api_core.exceptions import BadRequest, NotFound, Forbidden, GoogleAPIError

from typing import Union
import json


class BigQueryInputEngine(alteryx.AyxPluginBase):
    def __init__(self, n_tool_id: int, alteryx_engine: Sdk.AlteryxEngine, output_anchor_mgr: Sdk.OutputAnchorManager):
        super().__init__(n_tool_id, alteryx_engine, output_anchor_mgr)
        self.auth_type = None   # type: str
        self.service_file_loc = None # type: str
        self.service_account_json = None # type: str
        self.project_id = None   # type: str
        self.dataset_id = None # type: str
        self.table_id = None # type: str
        self.credentials = None # type: google.auth.credentials.Credentials
        self.record_limit = 0
        self.max_field_length = DEFAULT_MAX_FIELD_LENGTH
        self.using_custom_query = False
        self.custom_query = None
        self.sql_type = None
        self.max_repeated_records = 5

    def pi_add_incoming_connection(self, str_type: str, str_name: str) -> Union[alteryx.IncomingConnectionBase, None]:
        self.display_error_msg(
            self.alteryx_engine.xmsg("This tool does not accept incoming connections")
        )
        return None

    def init_end_user_credentials(self, str_xml: str):
        access_token_str = self.get_encrypted_xml_config_value(str_xml, 'accessToken')
        try:
            access_token = json.loads(access_token_str) if access_token_str else None
        except json.decoder.JSONDecodeError as e:
            self.invalidate_config(self.alteryx_engine.xmsg("Access token is not valid JSON"))
            return False
        if not access_token:
            self.invalidate_config(self.alteryx_engine.xmsg("No access token provided for end user authentication"))
            return False

        client_id = self.get_xml_config_value(str_xml, 'clientId')
        client_secret = self.get_encrypted_xml_config_value(str_xml, 'clientSecret')
        if not client_id:
            self.invalidate_config(self.alteryx_engine.xmsg('Please specify a Client ID for end user authentication'))
            return False
        if not client_secret:
            self.invalidate_config(
                self.alteryx_engine.xmsg('Please specify a Client secret for end user authentication')
            )
            return False

        try:
            self.credentials = credentials.Credentials(
                access_token['access_token'],
                refresh_token=access_token['refresh_token'],
                token_uri=OAUTH_AUD,
                client_id=client_id,
                client_secret=client_secret
            )
        except ValueError as e:
            self.invalidate_config(
                self.alteryx_engine.xmsg('Invalid access token or client credentials given for end user authentication')
            )
            return False
        except Exception as e:
            self.invalidate_config(self.alteryx_engine.xmsg('Error authenticating to Google BigQuery'))
            return False
        return True

    def init_service_principal_credentials(self, str_xml: str):
        json_str = self.get_encrypted_xml_config_value(str_xml, 'serviceCredsJSONStr')
        try:
            self.service_account_json = json.loads(json_str) if json_str else None
        except json.decoder.JSONDecodeError as e:
            self.invalidate_config(self.alteryx_engine.xmsg(
                "Invalid service account keyfile. Check that you specified a valid JSON keyfile."
            ))
            return False
        if not self.service_account_json:
            self.invalidate_config(self.alteryx_engine.xmsg(
                'Service account JSON not provided. Please specify a path to your service account file.'
            ))
            return False
        try:
            self.credentials = service_account.Credentials.from_service_account_info(
                self.service_account_json,
                scopes=[BIG_QUERY_SCOPE]
            )
        except ValueError as e:
            self.invalidate_config(self.alteryx_engine.xmsg("Invalid service account keyfile. @1", str(e)))
            return False

        return True

    def pi_init(self, str_xml: str):
        super().pi_init(str_xml)
        self.auth_type = self.get_xml_config_value(str_xml, 'authType')
        if self.auth_type == 'EndUser':
            if not self.init_end_user_credentials(str_xml):
                return
        elif self.auth_type == 'ServicePrincipal':
            if not self.init_service_principal_credentials(str_xml):
                return
        else:
            self.invalidate_config(self.alteryx_engine.xmsg('Invalid authentication mode selected: @1', self.auth_type))
            return

        selected_id = self.get_xml_config_value(str_xml, 'selectedId')
        self.using_custom_query = self.get_xml_config_value(str_xml, 'usingCustomFilter') == 'True'
        if not self.using_custom_query:
            try:
                self.project_id, remainder = selected_id.split(':')
                self.dataset_id, self.table_id = remainder.split('.')
            except (AttributeError, ValueError) as e:
                self.invalidate_config(self.alteryx_engine.xmsg('Invalid table selection: please select a table.'))
        else:
            self.custom_query = self.get_xml_config_value(str_xml, 'inputCustomFilter')
            if not self.custom_query:
                self.invalidate_config(self.alteryx_engine.xmsg('Please enter a non-empty custom query.'))
            self.project_id = self.get_xml_config_value(str_xml, 'customFilterProjectId')
            if not self.project_id:
                self.invalidate_config(self.alteryx_engine.xmsg('Please select a project to run your custom query against.'))

        self.sql_type = self.get_xml_config_value(str_xml, 'sqlType')

        try:
            self.max_repeated_records = int(self.get_xml_config_value(str_xml, 'inputRepeatedRecordLimit'))
        except TypeError:
            self.max_repeated_records = 0

        try:
            self.max_field_length = int(self.get_xml_config_value(str_xml, 'maxFieldLength'))
        except (ValueError, TypeError) as e:
            self.max_field_length = DEFAULT_MAX_FIELD_LENGTH

        should_limit_records = self.get_xml_config_value(str_xml, 'shouldLimitRecords')
        if should_limit_records == 'True':
            try:
                self.record_limit = int(self.get_xml_config_value(str_xml, 'inputRecordLimit'))
            except TypeError:
                self.record_limit = -1
        else:
            self.record_limit = -1

        should_infer_rr_limit = self.get_xml_config_value(str_xml, 'shouldInferRepeatedRecordLimit')
        if should_infer_rr_limit == 'True':
            self.log.debug('Inferring repeated record limit.')
            try:
                self.max_repeated_records = self.get_max_array_length()
            except (CannotInferRepeatedRecordLimit, Exception) as e:
                self.invalidate_config(self.alteryx_engine.xmsg('Could not infer repeated record limit.'))
                return

        metadata_dict = {
            'Host': BIG_QUERY_API_BASE,
            'ServiceName': 'BigQuery',
            'PluginName': 'BigQuery',
        }
        if self.using_custom_query:
            metadata_dict['Query'] = self.custom_query
        if not self.custom_query:
            metadata_dict['Database'] = self.dataset_id
            metadata_dict['Table'] = self.table_id
        self.alteryx_engine.create_connect_metadata(self.n_tool_id, metadata_dict)
        self.output_anchor = self.output_anchor_mgr.get_output_anchor('Output')  # type: Sdk.OutputAnchor

    def get_max_array_length(self):
        try:
            client = bigquery.Client(project=self.project_id, credentials=self.credentials)
            max_nested = 0
            dataset_ref = client.dataset(self.dataset_id)
            table_ref = dataset_ref.table(self.table_id)
            table = client.get_table(table_ref)
            query_shell = 'SELECT MAX(len) FROM (SELECT ARRAY_LENGTH({}) as len FROM {}.{}'
            if self.record_limit > 0:
                query_shell += ' LIMIT {}'.format(self.record_limit)
            query_shell += ')'
            for field in table.schema:
                if field.field_type == 'RECORD' and field.mode == 'REPEATED':
                    query = query_shell.format(field.name, self.dataset_id, self.table_id)
                    results = client.query(query).result()
                    max_nested_for_field = list(results)[0][0]
                    max_nested = max(max_nested, max_nested_for_field)
            return max_nested
        except BadRequest as e:
            self.log.exception(e)
            self.display_error_msg(
                self.alteryx_engine.xmsg("@1 Bad Request to Google API: @2", str(e.code.value), e.message)
            )
            raise CannotInferRepeatedRecordLimit('Bad Request')
        except NotFound as e:
            self.log.exception(e)
            self.display_error_msg(
                self.alteryx_engine.xmsg("@1 Bad Request to Google API: @2", str(e.code.value), e.message)
            )
            raise CannotInferRepeatedRecordLimit('Not Found')
        except Forbidden as e:
            self.log.exception(e)
            self.display_error_msg(self.alteryx_engine.xmsg("@1 Invalid Permissions: @2", str(e.code.value), e.message))
            raise CannotInferRepeatedRecordLimit('Forbidden')
        except GoogleAPIError as e:
            self.log.exception(e)
            self.display_error_msg(
                self.alteryx_engine.xmsg("@1 GoogleAPIError @2: @3", str(e.code.value), type(e).__name__, e.message)
            )
            raise CannotInferRepeatedRecordLimit('API Error')
        except Exception as e:
            self.log.exception('EXCEPTION in get_max_array_length: %s', str(e))
            raise CannotInferRepeatedRecordLimit('General')

    def pi_add_outgoing_connection(self, str_name: str) -> bool:
        return True

    def pi_push_all_records(self, n_record_limit: int = -1) -> bool:
        if not self.is_initialized:
            return False

        update_only = self.alteryx_engine.get_init_var(self.n_tool_id, 'UpdateOnly') == 'True' \
            or n_record_limit == 0
        record_limit = self.record_limit if self.record_limit > 0 else n_record_limit
        if record_limit == 0:
            update_only = True

        client = bigquery.Client(project=self.project_id, credentials=self.credentials)
        try:
            schema_reader, row_iterator = None, None
            if not self.using_custom_query:
                schema_reader, row_iterator, total_records = self._push_records_from_table(client, record_limit)
            else:
                schema_reader, row_iterator, total_records = self._push_records_from_query(client)

            record_info = schema_reader.get_record_info()
            self.output_anchor.init(record_info)

            if update_only:
                return True

            record_creator = record_info.construct_record_creator()  # type: Sdk.RecordCreator
            record_count = 0
            zero_total_records_warning_shown = False
            for result in row_iterator:
                if total_records == 0 and not zero_total_records_warning_shown:
                    self.display_warning_msg(
                        self.alteryx_engine.xmsg(
                            'Obtained unknown number of records from BigQuery. '
                            'Unable to estimate progress indicator while workflow is running.'
                        )
                    )
                    zero_total_records_warning_shown = True
                record_info_out = record_info.clone()  # type: Sdk.RecordInfo
                record_creator.reset()
                for field in row_iterator.schema:
                    for record_field, field_value in schema_reader.generate_records_from_field(result, field, record_info_out):
                        alteryx_helpers.set_value_on_field(record_field, record_creator, field_value, self.alteryx_engine)

                finalized_record = record_creator.finalize_record()
                self.output_anchor.push_record(finalized_record)
                record_count += 1
                if total_records != 0:
                    self.output_anchor.update_progress(record_count / total_records)

            finish_msg = self.alteryx_engine.xmsg(
                "@1 @2 records were read from @3:@4", str(record_count), self.table_id, self.project_id, self.dataset_id
            )
            if self.using_custom_query:
                finish_msg = self.alteryx_engine.xmsg(
                    "@1 records were read from a custom BigQuery Query.", str(record_count)
                )
            self.display_info_msg(finish_msg)

            self.output_anchor.update_progress(1.0)
            self.output_anchor.close()  # Close outgoing connections.
            return True
        except BadRequest as e:
            self.log.exception(e)
            self.display_error_msg(
                self.alteryx_engine.xmsg("@1 Bad Request to Google API: @2", str(e.code.value), e.message)
            )
        except NotFound as e:
            self.log.exception(e)
            self.display_error_msg(
                self.alteryx_engine.xmsg("@1 Bad Request to Google API: @2", str(e.code.value), e.message)
            )
        except Forbidden as e:
            self.log.exception(e)
            self.display_error_msg(
                self.alteryx_engine.xmsg("@1 Invalid Permissions: @2", str(e.code.value), e.message)
            )
        except GoogleAPIError as e:
            self.log.exception(e)
            self.display_error_msg(
                self.alteryx_engine.xmsg("@1 GoogleAPIError @2: @3", str(e.code.value), type(e).__name__, e.message)
            )
        except Exception as e:
            self.log.exception('EXCEPTION in pi_push_all_records: %s', e)
            self.display_error_msg(self.alteryx_engine.xmsg("There was an unexpected error."))
            return False

    def _push_records_from_query(self, client):
        job_config = bigquery.QueryJobConfig()
        job_config.use_legacy_sql = self.sql_type == 'LEGACY_SQL'
        # Data Query
        query_job = client.query(self.custom_query, job_config=job_config)
        query_result = query_job.result()
        # Count Query
        count_job_config = bigquery.QueryJobConfig()
        count_job_config.use_legacy_sql = self.sql_type == 'LEGACY_SQL'
        count_job = client.query('SELECT COUNT(*) FROM ({})'.format(self.custom_query.rstrip(';')), job_config=count_job_config)
        total_rows = list(count_job.result())[0][0]
        schema_reader = BigQuerySchemaReader(self.alteryx_engine, query_result, self.max_repeated_records, self.max_field_length)
        if schema_reader.has_repeated_records:
            total_rows *= self.max_repeated_records
        return schema_reader, query_result, total_rows

    def _push_records_from_table(self, client, record_limit):
        dataset_ref = client.dataset(self.dataset_id)
        table_ref = dataset_ref.table(self.table_id)
        table = client.get_table(table_ref)
        streaming_buffer = table.streaming_buffer
        total_rows = table.num_rows + streaming_buffer.estimated_rows if streaming_buffer else table.num_rows
        if record_limit > 0:
            record_limit = min(total_rows, record_limit)
        else:
            record_limit = total_rows
        schema_reader = BigQuerySchemaReader(self.alteryx_engine, table, self.max_repeated_records, self.max_field_length)
        row_iterator = client.list_rows(table, max_results=record_limit)
        return schema_reader, row_iterator, record_limit

    def pi_close(self, b_has_errors: bool):
        super().pi_close(b_has_errors)
        if self.output_anchor:
            self.output_anchor.assert_close()
