import AlteryxPythonSDK as Sdk
import logging
import xml.etree.ElementTree as Et
from abc import abstractmethod
from typing import Sequence, Union

from .helpers import build_record_info


class IncomingConnectionBase:
    """
    This optional class is returned by pi_add_incoming_connection, and it implements the incoming interface methods, to
    be utilized by the Alteryx engine to communicate with a plugin when processing an incoming connection.
    Prefixed with "ii", the Alteryx engine will expect the below four interface methods to be defined.
    """

    def __init__(self, parent: object):
        """
        Constructor for IncomingInterface.

        :param parent: AyxPlugin
        :type parent: AyxPluginBase
        """

        self.log = logging.getLogger(self.__class__.__name__)
        self.parent = parent
        self.is_initialized = False

    def ii_init(self, record_info_in: Sdk.RecordInfo) -> bool:
        """
        Called to report changes of the incoming connection's record metadata to the Alteryx engine.

        :param record_info_in: A RecordInfo object for the incoming connection's fields.
        """

        self.is_initialized = True
        return True

    @abstractmethod
    def ii_push_record(self, in_record: Sdk.RecordRef) -> bool:
        """
        Called when an input record is being sent to the plugin.

        :param in_record: The data for the incoming record.
        """

        pass

    @abstractmethod
    def ii_update_progress(self, d_percent: float):
        """
        Called by the upstream tool to report what percentage of records have been pushed.

        :param d_percent: Value between 0.0 and 1.0.
        """

        pass

    @abstractmethod
    def ii_close(self):
        """
        Called when the incoming connection has finished passing all of its records.
        """

        pass


class AyxPluginBase:
    """
    The Alteryx Engine plugin interface, to be utilized by the Alteryx engine to communicate with a plugin.
    Prefixed with "pi", the Alteryx engine will expect the below five interface methods to be defined.
    """

    @staticmethod
    def get_xml_config_value(str_xml: str, config_key: str) -> str:
        return Et.fromstring(str_xml).find(config_key).text if config_key in str_xml else None

    def __init__(self, n_tool_id: int, alteryx_engine: Sdk.AlteryxEngine, output_anchor_mgr: Sdk.OutputAnchorManager):
        self.log = logging.getLogger(self.__class__.__name__)
        # Default properties
        self.n_tool_id = n_tool_id
        self.alteryx_engine = alteryx_engine  # type: Sdk.AlteryxEngine
        self.output_anchor_mgr = output_anchor_mgr  # type: Sdk.OutputAnchorManager
        self.output_anchor = None  # type: Sdk.OutputAnchor

        self.is_initialized = False

    def pi_init(self, str_xml: str):
        """
        Handles extracting user-entered file name, and input data verification.
        Called when the Alteryx engine is ready to provide the tool configuration from the GUI.

        :param str_xml: The raw XML from the GUI.
        """

        self.is_initialized = True

    def pi_add_incoming_connection(self, str_type: str, str_name: str) -> Union[IncomingConnectionBase, None]:
        """
        The IncomingInterface objects are instantiated here, one object per incoming connection, however since
        this tool does not accept an incoming connection, instantiation is not needed and "ii" methods won't be called.
        Called when the Alteryx engine is attempting to add an incoming data connection.

        :param str_type: The name of the input connection anchor, defined in the Config.xml file.
        :param str_name: The name of the wire, defined by the workflow author.
        :return: self.
        """

        return None

    def pi_add_outgoing_connection(self, str_name: str) -> bool:
        """
        Called when the Alteryx engine is attempting to add an outgoing data connection.

        :param str_name: The name of the output connection anchor, defined in the Config.xml file.
        :return: True signifies that the connection is accepted.
        """

        return False

    @abstractmethod
    def pi_push_all_records(self, n_record_limit: int) -> bool:
        """
        Handles pushing records out to downstream tool(s).
        Called when a tool has no incoming data connection.

        :param n_record_limit: Set it to <0 for no limit, 0 for no records, and >0 to specify the number of records.
        :return: False if there are issues with the input data or if the workflow isn't being ran, otherwise True.
        """

        pass

    def pi_close(self, b_has_errors: bool):
        """
        Called after all records have been processed.

        :param b_has_errors: Set to true to not do the final processing.
        """

        pass

    def display_error_msg(self, msg_string: str):
        """
        A non-interface method, that is responsible for displaying the relevant ERROR message in Designer.

        :param msg_string: The custom error message.
        """

        self.log.error(msg_string)
        self.alteryx_engine.output_message(self.n_tool_id, Sdk.EngineMessageType.error, msg_string)

    def display_warning_msg(self, msg_string: str):
        """
        A non-interface method, that is responsible for displaying the relevant WARNING message in Designer.

        :param msg_string: The custom error message.
        """

        self.log.warning(msg_string)
        self.alteryx_engine.output_message(self.n_tool_id, Sdk.EngineMessageType.warning, msg_string)

    def display_info_msg(self, msg_string: str):
        """
        Display an informational message to the user

        :param msg_string: The message to display
        """
        self.log.info(msg_string)
        self.alteryx_engine.output_message(self.n_tool_id, Sdk.EngineMessageType.info, msg_string)

    def invalidate_config(self, msg: str):
        self.log.warning("Tool configuration is not valid: {}".format(msg))
        self.is_initialized = False
        self.display_error_msg(msg)

    def build_record_info(self, record_fields: Sequence[dict], get_field_source=lambda field: '') -> Sdk.RecordInfo:
        """
        A non-interface helper for pi_push_all_records() responsible for creating the outgoing record layout.

        :param get_field_source: A function to call that will return a string containing the field source information
        :param record_fields: The name for csv file reader.
        :return: The outgoing record layout, otherwise nothing.
        """

        return build_record_info(self.alteryx_engine, record_fields, get_field_source)

    def get_encrypted_xml_config_value(self, xml: str, config_key: str):
        """
        Retrieve and decrypt a value from the XML configuration string that was encrypted by the
        configuration GUI.

        :param xml: XML Configuration
        :param config_key: The name of the configuration value to retrieve and decrypt
        :return: The decrypted configuration value if it exists and can be decrypted; otherwise None
        """
        encrypted = self.get_xml_config_value(xml, config_key)

        if encrypted is None:
            return None
        try:
            return self.alteryx_engine.decrypt_password(encrypted)
        except Exception as e:
            self.log.exception("Failed decrypting value for {}".format(config_key), exc_info=e)
            self.display_error_msg(self.alteryx_engine.xmsg("Failed to decrypt value for @1!", config_key))
            return None
