import csv
import io

from typing import IO, Iterator, Union, Callable

import AlteryxPythonSDK as Sdk

from connectors_core.alteryx.helpers import build_record_info
from .interfaces import FileInputReader
from .errors import DelimitedIndexError


class DelimitedReader(FileInputReader):

    def __init__(
            self,
            stream: IO,
            output_anchor: Sdk.OutputAnchor,
            alteryx_engine: Sdk.AlteryxEngine,
            stream_length_bytes: Union[int, None] = None,
            field_source_resolver: Callable[[Sdk.Field], str] = lambda f: '',
            delimiter: str=',',
            quotechar: str='"',
            encoding: str='utf-8',
            has_header_row: bool=True,
            binary_input=False,
            max_field_size=255,
            start_data_at_line=1
    ):
        super().__init__(stream, output_anchor, alteryx_engine, stream_length_bytes, field_source_resolver)
        self.encoding = encoding
        self.quotechar = quotechar
        self.start_data_at_line = start_data_at_line
        self.has_header_row = has_header_row
        self.max_field_size = max_field_size
        self.delimiter = delimiter
        self.csv_reader = None
        self.binary_input = binary_input
        self.csv_reader = csv.reader(
            self.bytes_to_str(self.stream, self.encoding) if self.binary_input else self.stream,
            delimiter=self.delimiter,
            quotechar=self.quotechar
        )  # type: Iterator[list]
        self.first_row = None

    def init(self) -> Sdk.RecordInfo:
        record = next(self.csv_reader)
        if self.has_header_row:
            self.record_info = build_record_info(
                self.alteryx_engine,
                [{
                    'name': r,
                    'type': Sdk.FieldType.v_wstring,
                    'size': self.max_field_size,
                    'scale': 0,
                    'description': None
                } for r in record],
                self.field_source_resolver
            )
        else:
            self.record_info = build_record_info(
                self.alteryx_engine,
                [{
                    'name': "Field_{}".format(i),
                    'type': Sdk.FieldType.v_wstring,
                    'size': self.max_field_size,
                    'scale': 0,
                    'description': None
                } for i, r in enumerate(record)],
                self.field_source_resolver
            )
            self.first_row = record

        self.output_anchor.init(self.record_info)
        self.record_creator = self.record_info.construct_record_creator()
        self.is_initialized = True
        return self.record_info

    def read(self, n_record_limit=-1) -> int:
        record_count = 0
        rows_seen = 0
        if not self.has_header_row and self.start_data_at_line == 1 and self.first_row is not None:
            self.push_record(self.first_row)
            record_count += 1
        elif not self.has_header_row and self.start_data_at_line > 1:
            rows_seen += 1

        for record_index, record in enumerate(self.csv_reader):
            rows_seen += 1
            if self.start_data_at_line > rows_seen:
                continue
            self.push_record(record)
            record_count = record_count + 1
            if 0 < n_record_limit <= record_count:
                break

        return record_count

    def push_record(self, record):
        record_info_out = self.record_info.clone()
        self.record_creator.reset()
        try:
            for field_index, field_key in enumerate(record):
                record_info_out[field_index].set_from_string(self.record_creator, field_key)
        except IndexError as e:
            raise DelimitedIndexError(e) from e
        finalized_record_out = self.record_creator.finalize_record()
        self.output_anchor.push_record(finalized_record_out, no_auto_close=False)
        self.report_input_stream_progress()

    @staticmethod
    def bytes_to_str(f: io.BufferedIOBase, encoding: str = 'utf-8'):
        text_wrapper = io.TextIOWrapper(f, encoding=encoding)
        for line in text_wrapper:
            yield line
