from onedrivesdk import AuthProvider
from onedrivesdk.version_bridge.fragment_upload import ItemUploadFragmentBuilder, ItemUploadFragment
from onedrivesdk.options import *
from onedrivesdk.model.item import Item
from onedrivesdk.request.item_create_session import ItemCreateSessionRequestBuilder
from onedrivesdk.request_base import RequestBase
from onedrivesdk.helpers.file_slice import FileSlice
from connectors_core.file import FileExistsBehavior
import AlteryxPythonSDK as Sdk
import json
import requests
import os
import logging

AUTH_SERVER_URL = "https://login.microsoftonline.com/common/oauth2/v2.0/authorize"
AUTH_TOKEN_URL = "https://login.microsoftonline.com/common/oauth2/v2.0/token"


class OneDriveAuthProvider(AuthProvider):
    def __init__(self, http_provider, access_token, refresh_token, client_secret, redirect_uri, client_id, scopes, expires_in, alteryx_engine: Sdk.AlteryxEngine = None):
        super().__init__(http_provider=http_provider, client_id=client_id, scopes=scopes, auth_server_url=AUTH_SERVER_URL, auth_token_url=AUTH_TOKEN_URL, session_type=None)

        self._session = self._session_type(
            "code",
            expires_in,
            scopes,
            access_token,
            client_id,
            AUTH_SERVER_URL,
            redirect_uri,
            refresh_token,
            client_secret
        )

        self.alteryx_engine = alteryx_engine

    def authenticate_request(self, request):
        if self._session is None:
            if not self.alteryx_engine:
                raise RuntimeError("""Session must be authenticated before applying authentication to a request""")
            else:
                raise RuntimeError(self.alteryx_engine.xmsg("""Session must be authenticated before applying authentication to a request"""))

        if self._session.is_expired():
            self.refresh_token()

        request.append_option(
            HeaderOption("Authorization",
                         "bearer {}".format(self._session.access_token)))

    def refresh_token(self):
        """Refresh the token currently used by the session"""
        if self._session is None:
            if not self.alteryx_engine:
                raise RuntimeError("""Session must be authenticated 
                    before refreshing token.""")
            else:
                raise RuntimeError(self.alteryx_engine.xmsg("""Session must be authenticated 
                    before refreshing token."""))

        if self._session.refresh_token is None:
            if not self.alteryx_engine:
                raise RuntimeError("""Refresh token not present.""")
            else:
                raise RuntimeError(self.alteryx_engine.xmsg("""Refresh token not present."""))

        params = {
            "refresh_token": self._session.refresh_token,
            "client_id": self._session.client_id,
            "redirect_uri": self._session.redirect_uri,
            "grant_type": "refresh_token"
        }

        if self._session.client_secret is not None:
            params["client_secret"] = self._session.client_secret

        headers = {"Content-Type": "application/x-www-form-urlencoded"}
        resp = requests.post(
            'https://login.microsoftonline.com/common/oauth2/v2.0/token',
            headers=headers,
            data=params
        )
        rcont = json.loads(resp.content)
        self._session.refresh_session(rcont["expires_in"],
                                      rcont["scope"],
                                      rcont["access_token"],
                                      rcont["refresh_token"])


class OneDriveFragmentUploadBuilder(RequestBase):
    def __init__(self, request_url, client, content_local_path, conflict_behavior):
        super().__init__(request_url, client, options=None)
        self._file_handle = open(content_local_path, 'rb')
        self._total_length = os.stat(content_local_path).st_size
        self.session = None
        self.conflict_behavior = conflict_behavior
        self._request_url = None
        self.url = request_url
        self.log = logging.getLogger(self.__class__.__name__)

    def request(self, begin, length, options=None):
        opts = None
        if not (options is None or len(options) == 0):
            opts = options.copy()
        else:
            opts = []

        if (self.conflict_behavior == FileExistsBehavior.Overwrite) or (self.conflict_behavior == 'replace'):
            self.conflict_behavior = 'replace'
        elif self.conflict_behavior == FileExistsBehavior.Append or (self.conflict_behavior == 'rename'):
            self.conflict_behavior = 'rename'
        else:
            self.conflict_behavior = 'fail'

        if not self.session:
            item = Item({'@name.conflictBehavior': self.conflict_behavior})
            self.session = self.create_session(item).post()
            self._request_url = self.session.upload_url

        opts.append(HeaderOption("Content-Type", "application/octet-string"))
        opts.append(HeaderOption("Content-Range", "bytes %d-%d/%d" % (begin, begin + length - 1, self._total_length)))
        opts.append(HeaderOption("Content-Length", str(length)))
        file_slice = FileSlice(self._file_handle, begin, length=length)

        req = ItemUploadFragment(self._request_url, self._client, opts, file_slice)
        return req

    def post(self, begin, length, options=None):
        return self.request(begin, length, options).post()

    def create_session(self, item=None):
        return ItemCreateSessionRequestBuilder(self.append_to_request_url("createUploadSession"), self._client, item=item)

    def append_to_request_url(self, url_segment):
        return self.url + "/" + url_segment

