import AlteryxPythonSDK as Sdk
import connectors_core.alteryx.helpers as alteryx_helpers
from connectors_core import alteryx
from connectors_core.bigquery.schemas import BigQuerySchemaWriter, SchemaConversionException

from google.cloud import bigquery
from google.api_core.exceptions import NotFound, Forbidden, GoogleAPIError


class BigQueryIncomingConnectionsHandler(alteryx.IncomingConnectionBase):

    def __init__(self, parent: alteryx.AyxPluginBase):
        super().__init__(parent)
        self.record_info = None  # type: Sdk.RecordInfo
        self.client = None
        self.dataset = None
        self.table = None
        self.schema_writer = None
        self.row_buffer = []
        self.row_insert_batch_size = None
        self.is_initialized = False

    def get_client(self):
        if not self.client:
            self.client = bigquery.Client(project=self.parent.project_id, credentials=self.parent.credentials)
        return self.client

    def init_dataset(self):
        if self.dataset is not None:
            return
        client = self.get_client()
        dataset_ref = client.dataset(self.parent.dataset_id)
        self.dataset = client.get_dataset(dataset_ref)

    def init_table(self):
        if self.table is not None:
            return
        if self.dataset is None:
            self.init_dataset()

        client = self.get_client()
        table_ref = self.dataset.table(self.parent.table_id)
        self.table = client.get_table(table_ref)

    def ii_init(self, record_info_in: Sdk.RecordInfo) -> bool:
        if not self.parent.is_initialized:
            return False
        self.record_info = record_info_in
        try:
            self.init_dataset()
            self.init_table()
        except NotFound as e:
            self.parent.display_error_msg(self.parent.alteryx_engine.xmsg(
                "@1 Bad Request to Google API: @2", str(e.code.value), e.message
            ))
            return False
        except Forbidden as e:
            self.parent.display_error_msg(self.parent.alteryx_engine.xmsg(
                "@1 Invalid Permissions: @2", str(e.code.value), e.message
            ))
        except GoogleAPIError as e:
            self.parent.display_error_msg(self.parent.alteryx_engine.xmsg(
                "@1 GoogleAPIError @2: @3", str(e.code.value), type(e).__name__, e.message
            ))
        self.schema_writer = BigQuerySchemaWriter(
            self.table, self.parent.max_repeated_records, self.parent.max_field_length, self.record_info
        )
        self.row_insert_batch_size = self.parent.row_insert_batch_size
        self.is_initialized = True
        return True

    def ii_push_record(self, in_record: Sdk.RecordRef) -> bool:
        try:
            row = self.schema_writer.build_row_from_record(in_record, self.parent.alteryx_engine)
        except SchemaConversionException as e:
            self.parent.display_error_msg(e.message)
            return False
        self.row_buffer.append(row)
        if len(self.row_buffer) >= self.row_insert_batch_size:
            buffer_slice = self.row_buffer[:self.row_insert_batch_size]
            try:
                errors = self.get_client().insert_rows(self.table, buffer_slice)
                self.row_buffer = self.row_buffer[self.row_insert_batch_size:]
                if errors:
                    self.parent.display_error_msg(errors)
            except Forbidden as e:
                self.parent.display_error_msg(self.parent.alteryx_engine.xmsg(
                    "@1 Invalid Permissions: @2", str(e.code.value), e.message
                ))
            except GoogleAPIError as e:
                self.parent.display_error_msg(self.parent.alteryx_engine.xmsg(
                    "@1 GoogleAPIError @2: @3", str(e.code.value), type(e).__name__, e.message
                ))
        return True

    def ii_update_progress(self, d_percent: float):
        try:
            self.parent.alteryx_engine.output_tool_progress(self.parent.n_tool_id, d_percent)
        except AttributeError:
            pass

    def ii_close(self):
        try:
            if self.row_buffer:
                errors = self.get_client().insert_rows(self.table, self.row_buffer)
                if errors and self.parent.is_initialized:
                    self.parent.display_error_msg(errors)
        except Forbidden as e:
            if self.parent:
                self.parent.display_error_msg(self.parent.alteryx_engine.xmsg(
                    "@1 Invalid Permissions: @2", str(e.code.value), e.message
                ))
        except GoogleAPIError as e:
            if self.parent:
                self.parent.display_error_msg(self.parent.alteryx_engine.xmsg(
                    "@1 GoogleAPIError @2: @3", str(e.code.value), type(e).__name__, e.message
                ))
        except AttributeError:
            pass
        self.is_initialized = False
