import csv
import json
import io
import os

import fastavro.write as avro

import onedrivesdk as od
import AlteryxPythonSDK as Sdk

import connectors_core.file.errors as fileerrors
import connectors_core.file.interfaces as interfaces
import connectors_core.file.avro as avro_helper
import connectors_core.alteryx.helpers as alteryx_helpers
from connectors_core.onedrive.od import OneDriveFragmentUploadBuilder

from abc import abstractmethod
from typing import IO, Callable, Any

from connectors_core.file import FileExistsBehavior


class ODFileWriterBase(interfaces.FileWriter):

    def __init__(
            self,
            client: od.OneDriveClient,
            file_path: str,
            temp_file: str,
            file_exists_behavior: FileExistsBehavior,
            data_conversion_error_handler: Callable[[Sdk.Field, Sdk.RecordRef, BaseException], Any] = None,
            alteryx_engine: Sdk.AlteryxEngine = None
    ):
        super().__init__()
        self.file_exists_behavior = file_exists_behavior
        self.data_conversion_handler = data_conversion_error_handler
        self.client = client
        self.file_path = file_path
        self.temp_file = temp_file
        self.stream = None
        self._is_open = False
        self.records_written = 0
        self.record_info = None
        self.session = None
        self.options = None
        self.item_request = None
        self.url = None
        self.alteryx_engine = alteryx_engine

    def init(self, record_info: Sdk.RecordInfo):
        self.record_info = record_info

        self.url = 'https://graph.microsoft.com/v1.0/me/drive/items/root:/%s:' % self.file_path

    def open(self):
        self.stream = io.open(self.temp_file, "wb")
        self._is_open = True

    def write(self, record: Sdk.RecordRef):
        if not self.stream:
            if not self.alteryx_engine:
                raise fileerrors.FileOpenError("open() must be called on the writer before records are written")
            else:
                raise fileerrors.FileOpenError(self.alteryx_engine.xmsg("open() must be called on the writer before records are written"))
        self.stream.write(self.get_bytes(record))
        self.stream.flush()
        self.records_written += 1
        return True

    def flush(self):
        if self.stream:
            self.stream.flush()

    def close(self):
        self.flush()
        if self.stream:
            self.stream.close()
            self.stream = None
            uploader = OneDriveFragmentUploadBuilder(request_url=self.url, client=self.client, content_local_path=self.temp_file, conflict_behavior=self.file_exists_behavior)
            begin = 0
            length = 6200
            file_size = os.stat(self.temp_file).st_size
            while True:
                if begin+length > file_size:
                    length = file_size - begin
                    uploader.post(begin, length)
                    break
                uploader.post(begin, length)
                begin += length
        if self.client:
            self.client = None

    def is_open(self):
        return self._is_open

    # def file_exists(self):
    #     return False

    @abstractmethod
    def get_bytes(self, record: Sdk.RecordRef) -> bytes:
        pass


class ODDelimitedFileWriter(ODFileWriterBase):

    def __init__(
            self,
            client: od.OneDriveClient = None,
            file_path: str = None,
            temp_file: str = None,
            write_header: bool = True,
            delimiter: str = ',',
            data_conversion_error_handler: Callable[[Sdk.Field, Sdk.RecordRef, BaseException], Any] = None,
            file_exists_behavior: FileExistsBehavior=FileExistsBehavior.Abort.name,
            alteryx_engine: Sdk.AlteryxEngine = None,
            encoding: str = None
    ):
        super().__init__(client, file_path=file_path, temp_file=temp_file, data_conversion_error_handler=data_conversion_error_handler, file_exists_behavior=file_exists_behavior)
        self.delimiter = delimiter
        self.write_header = write_header
        self.encoding = encoding

    def init(self, record_info: Sdk.RecordInfo):
        super().init(record_info)

    def write(self, record: Sdk.RecordInfo):
        if self.write_header and self.records_written == 0:
            self.write_header_row()
        return super().write(record)

    def get_bytes(self, record: Sdk.RecordRef) -> bytes:
        values = [
            alteryx_helpers.get_value_for_field(field, record, self.data_conversion_handler)
            for field in self.record_info
        ]
        with io.StringIO('') as temp_stream:
            self.get_csv_writer(temp_stream).writerow(values)
            return temp_stream.getvalue().encode(self.encoding)

    def write_header_row(self):
        field_names = []
        for field in self.record_info:
            field_names.append(field.name)
        with io.StringIO('') as temp_stream:
            self.get_csv_writer(temp_stream).writerow(field_names)
            self.stream.write(temp_stream.getvalue().encode(self.encoding))

    def get_csv_writer(self, stream: IO) -> csv.writer:
        return csv.writer(stream, delimiter=self.delimiter)


class ODAvroFileWriter(ODFileWriterBase):

    def __init__(
            self,
            client: od.OneDriveClient,
            file_path: str,
            temp_file: str,
            data_conversion_error_handler: Callable[[Sdk.Field, Sdk.RecordRef, BaseException], Any] = None,
            file_exists_behavior: FileExistsBehavior=FileExistsBehavior.Abort,
            alteryx_engine: Sdk.AlteryxEngine = None
    ):
        super().__init__(client,
                         file_path=file_path,
                         temp_file=temp_file,
                         data_conversion_error_handler=data_conversion_error_handler,
                         file_exists_behavior=file_exists_behavior,
                         alteryx_engine=alteryx_engine)
        self.schema = None
        self.writer = None

    def init(self, record_info: Sdk.RecordInfo):
        super().init(record_info)
        self.schema = {
            'doc': 'Alteryx One Drive File Output',
            'name': 'AlteryxODOutput',
            'namespace': 'alteryx.od',
            'type': 'record',
            'fields': [
                {'name': f.name, 'type': avro_helper.get_avro_type_for_sdk_type(f.type, self.alteryx_engine)}
                for f in record_info
            ]
        }

    def open(self):
        super().open()
        self.writer = avro.Writer(self.stream, self.schema)

    def flush(self):
        if self.writer:
            self.writer.dump()
        super().flush()

    def close(self):
        super().close()
        self.writer = None

    def write(self, record: Sdk.RecordRef):
        self.writer.write(
            {
                f.name:
                    avro_helper.get_avro_value_from_field(f, record, self.data_conversion_handler, self.alteryx_engine)
                    for f in self.record_info
            }
        )
        return True

    def get_bytes(self, record: Sdk.RecordRef) -> bytes:
        raise NotImplementedError()


class ODJSONFileWriter(ODFileWriterBase):
    def __init__(
            self,
            client: od.OneDriveClient,
            file_path: str,
            temp_file: str,
            data_conversion_error_handler: Callable[[Sdk.Field, Sdk.RecordRef, BaseException], Any] = None,
            file_exists_behavior: FileExistsBehavior=FileExistsBehavior.Abort,
            alteryx_engine: Sdk.AlteryxEngine = None
    ):
        super().__init__(client, file_path=file_path, temp_file=temp_file, data_conversion_error_handler=data_conversion_error_handler, file_exists_behavior=file_exists_behavior)

    def open(self):
        super().open()
        self.stream.write(b'[\n')

    def write(self, record: Sdk.RecordRef):
        if self.records_written > 0:
            self.stream.write(b',\n')
        return super().write(record)

    def get_bytes(self, record: Sdk.RecordRef) -> bytes:
        return json.dumps(
            {
                f.name:
                    alteryx_helpers.get_value_for_field(f, record, self.data_conversion_handler)
                    for f in self.record_info
            }
        ).encode()

    def close(self):
        if self.stream:
            self.stream.write(b']')
        super().close()
